/*
 * Copyright (c) 2008 CACE Technologies, Davis (California)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the CACE Technologies nor the names of its 
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

//
// in order to use the old scanf
//
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <conio.h>

#define WPCAP
#define HAVE_REMOTE
#include <pcap.h>

#include <airpcap.h>
#include "PpiHeader.h"

static
VOID DissectPacket(struct pcap_pkthdr *pPcapHeader, const u_char *pData);

int main(int argc, char* argv[])
{
	int result;
	//
	// error message from pcap_findalldevs
	//
	char errBuf[PCAP_ERRBUF_SIZE];
	//
	// head of the list of adapters
	//
	pcap_if_t *pIfList;
	//
	// to scan the list of adapters
	//
	pcap_if_t *pIfCursor;
	pcap_t *pPcap;
	ULONG numAdapters;
	ULONG adIndex;
	ULONG i;
	const u_char *pData;
	struct pcap_pkthdr *pPcapHeader;
	PAirpcapHandle pAirpcap;
	AirpcapChannelInfo channelInfo;

	(VOID)(argc);
	(VOID)(argv);

	result = pcap_findalldevs(&pIfList, errBuf);

	if (result != 0)
	{
		printf("There was an error listing the adapters, %s\n", errBuf);
		return -1;
	}

	numAdapters = 0;
	for (pIfCursor = pIfList; pIfCursor != NULL; pIfCursor = pIfCursor->next)
	{
		printf("%u\n", numAdapters);
		printf("Name = %s\n", pIfCursor->name);
		printf("Description = %s\n", pIfCursor->description);
		printf("\n");
		numAdapters ++;
	}

	//
	// loop until the user chooses a valid adapter index
	//
	do
	{
		printf("Adapter (0 - %u)? ", numAdapters - 1);
	}
	while(scanf("%u", &adIndex) == 0 || adIndex >= numAdapters);

	//
	// pick the adapter from the list
	//
	for (pIfCursor = pIfList, i = 0; i != adIndex; pIfCursor = pIfCursor->next, i++);

	//
	// open the chosen adapter
	//
	pPcap = pcap_open_live(pIfCursor->name, 65535, 1, 1000, errBuf);

	//
	// we do not need the adapter list any more, free it
	//
	pcap_freealldevs(pIfList);

	if (pPcap == NULL)
	{
		printf("Error opening adapter %s\n", errBuf);
		return -1;
	}

	printf("Successfully opened adapter\n");

	//
	// check if we opened an airpcap adapter by querying the airpcap handle
	//
	pAirpcap = pcap_get_airpcap_handle(pPcap);

	if (pAirpcap == NULL)
	{
		printf("It's not an AirPcap adapter\n");
		pcap_close(pPcap);
		return -1;
	}

	//
	// set the airpcap link type to PPI
	//
	if (AirpcapSetLinkType(pAirpcap, AIRPCAP_LT_802_11_PLUS_PPI) == FALSE)
	{
		printf("Error setting the Airpcap link type to PPI: %s\n", AirpcapGetLastError(pAirpcap));
		pcap_close(pPcap);
		return -1;
	}

	channelInfo.Frequency = 2437; // 2437 = channel 6 (B/G)
	channelInfo.ExtChannel = 0;	
	channelInfo.Flags = 0;
	channelInfo.Reserved[0] = 0;
	channelInfo.Reserved[1] = 0;

	if (AirpcapSetDeviceChannelEx(pAirpcap, channelInfo) == FALSE)
	{
		printf("Error setting the channel: %s\n", AirpcapGetLastError(pAirpcap));
		pcap_close(pPcap);
		return -1;
	}

	//
	//receive the packets and process them
	//
	do
	{
		if (_kbhit())
		{
			//
			// the user wants to exit
			//
			// NOTE: be careful not to use this approach in production code.
			// kbhit() is an expensive operation if called for every received packet.
			//
			// call it
			// every X received packets or
			// from within the if (result == 0){} code i.e. if a timeout occurs
			//
			break;
		}

		result = pcap_next_ex(pPcap, &pPcapHeader, &pData);

		if (result < 0)
		{
			printf("Error receiving packets: %s\n", pcap_geterr(pPcap));
			//
			// exit from the loop
			//
			break;
		}

		if (result == 0)
		{
			//
			// timeout elapsed and no packets where received
			//
			continue;
		}

		//
		// we have a valid packet, try to extract the mac address and check if it's IP
		//
		DissectPacket(pPcapHeader, pData);

	}while(TRUE);

	pcap_close(pPcap);

	return 0;
}

//
// this routine assumes that the packet link type is PPI
//
static
VOID DissectPacket(struct pcap_pkthdr *pPcapHeader, const u_char *pData)
{
	PPI_PACKET_HEADER *pPpiHeader;
	BYTE *pDot11Data;
	USHORT	frameControl;

	//
	// first let's extract the PPI header and skip it
	//
	if (pPcapHeader->caplen < sizeof(PPI_PACKET_HEADER))
	{
		return;
	}

	pPpiHeader = (PPI_PACKET_HEADER*)pData;

	//
	// let's skip the PPI header and go to the data
	//
	if (pPcapHeader->caplen < pPpiHeader->PphLength)
	{
		printf("Truncated PPI header\n");
		return ;
	}

	pDot11Data = (BYTE*)(pData + pPpiHeader->PphLength);	

	//
	// let's extract the FrameControl
	//
	if (pPcapHeader->caplen < (ULONG)pPpiHeader->PphLength + 2)
	{
		printf("Truncated 802.11 frame\n");
		return;
	}

	//
	// the frame control is saved in little endian byte order, not
	// network byte order. we do not need to swap the bytes on x86
	//
	frameControl = *(USHORT*)pDot11Data;


	//
	// bits 2-3 are the frame type
	//
	BYTE type = (BYTE)((frameControl >> 2) & 0x03);

	//
	// bits 4-7 are the frame type
	//
	BYTE subType = (BYTE)((frameControl >> 4) & 0x0F);

	//
	// print the type and subtype of frame that we have received
	//
	switch(type)
	{
	case 0x0:	
		printf("MANAGEMENT frame: ");
		switch(subType)
		{
		case 0x00:	printf("Association request\n");	break;
		case 0x01:	printf("Association response\n");	break;
		case 0x02:	printf("Reassociation request\n");	break;
		case 0x03:	printf("Reassociation response\n");	break;
		case 0x04:	printf("Probe request\n");			break;
		case 0x05:	printf("Probe response\n");			break;
		case 0x08:	printf("Beacon\n");					break;
		default:	printf("Unknown (%02x)\n", subType);
		}
		break;

	case 0x1:	
		printf("CONTROL frame: ");
		switch(subType)
		{
		case 0x0A:	printf("PS-Poll\n");	break;
		case 0x0B:	printf("RTS\n");	break;
		case 0x0C:	printf("CTS\n");	break;
		case 0x0D:	printf("ACK\n");	break;
		case 0x0E:	printf("CF End\n");			break;
		case 0x0F:	printf("CD End + Probe response\n");			break;
		default:	printf("Unknown (%02x)\n", subType);
		}
		break;

	case 0x2:	
		printf("DATA frame: ");
		switch(subType)
		{
		case 0x00:	printf("DATA\n");	break;
		case 0x01:	printf("DATA+CF-Ack\n");	break;
		case 0x02:	printf("DATA+CF-Poll\n");	break;
		case 0x03:	printf("DATA+CF-Poll+CF-Ack\n");	break;
		case 0x04:	printf("NULL DATA\n");	break;
		case 0x05:	printf("NULL DATA+CF-Ack\n");	break;
		case 0x06:	printf("NULL DATA+CF-Poll\n");	break;
		case 0x07:	printf("NULL DATA+CF-Poll+CF-Ack\n");	break;
		case 0x08:	printf("QoS-DATA\n");	break;
		case 0x09:	printf("QoS-DATA+CF-Ack\n");	break;
		case 0x0a:	printf("QoS-DATA+CF-Poll\n");	break;
		case 0x0b:	printf("QoS-DATA+CF-Poll+CF-Ack\n");	break;
		case 0x0c:	printf("NULL QoS-DATA\n");	break;
		case 0x0d:	printf("NULL QoS-DATA+CF-Ack\n");	break;
		case 0x0e:	printf("NULL QoS-DATA+CF-Poll\n");	break;
		case 0x0f:	printf("NULL QoS-DATA+CF-Poll+CF-Ack\n");	break;
		}
		break;

	default: printf("Unkwown type (%02x)\n", type);
	}

}