/*
 * Copyright (c) 2008 CACE Technologies, Davis (California)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the CACE Technologies nor the names of its 
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

//
// in order to use the old scanf
//
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <conio.h>

#define WPCAP
#define HAVE_REMOTE
#include <pcap.h>


typedef struct _IP_HEADER
{
	BYTE	VerIHL;	
	BYTE	TOS;
	USHORT	TotalLength;
	USHORT	Identification;
	USHORT	FlagsFragmentOffset;
	BYTE	TTL;
	BYTE	Protocol;
	USHORT	HeaderChecksum;
	ULONG	SourceAddress;
	ULONG	DestAddress;
}
	IP_HEADER, *PIP_HEADER;

C_ASSERT(sizeof(IP_HEADER) == 20);

static
VOID DissectPacket(struct pcap_pkthdr *pPcapHeader, const u_char *pData);

int main(int argc, char* argv[])
{
	int result;
	//
	// error message from pcap_findalldevs
	//
	char errBuf[PCAP_ERRBUF_SIZE];
	//
	// head of the list of adapters
	//
	pcap_if_t *pIfList;
	//
	// to scan the list of adapters
	//
	pcap_if_t *pIfCursor;
	pcap_t *pPcap;
	ULONG numAdapters;
	ULONG adIndex;
	ULONG i;
	int datalink;
	const u_char *pData;
	struct pcap_pkthdr *pPcapHeader;
	struct bpf_program bpfProgram;

	(VOID)(argc);
	(VOID)(argv);

	result = pcap_findalldevs(&pIfList, errBuf);

	if (result != 0)
	{
		printf("There was an error listing the adapters, %s\n", errBuf);
		return -1;
	}

	numAdapters = 0;
	for (pIfCursor = pIfList; pIfCursor != NULL; pIfCursor = pIfCursor->next)
	{
		printf("%u\n", numAdapters);
		printf("Name = %s\n", pIfCursor->name);
		printf("Description = %s\n", pIfCursor->description);
		printf("\n");
		numAdapters ++;
	}

	//
	// loop until the user chooses a valid adapter index
	//
	do
	{
		printf("Adapter (0 - %u)? ", numAdapters - 1);
	}
	while(scanf("%u", &adIndex) == 0 || adIndex >= numAdapters);

	//
	// pick the adapter from the list
	//
	for (pIfCursor = pIfList, i = 0; i != adIndex; pIfCursor = pIfCursor->next, i++);

	//
	// open the chosen adapter
	//
	pPcap = pcap_open_live(pIfCursor->name, 65535, 1, 1000, errBuf);

	//
	// we do not need the adapter list any more, free it
	//
	pcap_freealldevs(pIfList);

	if (pPcap == NULL)
	{
		printf("Error opening adapter %s\n", errBuf);
		return -1;
	}

	printf("Successfully opened adapter\n");

	//
	// first, check that the link type is ethernet, i.e. DLT_EN10MB
	//
	datalink = pcap_datalink(pPcap);

	if (datalink != DLT_EN10MB)
	{
		printf("This application can decode ethernet packets only, data link type %d is not supported\n", datalink);
		pcap_close(pPcap);
		return -1;
	}

	//
	// let's compile a filter for ICMP packets
	//
	result = pcap_compile(
		pPcap, 
		&bpfProgram,	// the resulting BPF filter code
		"icmp",			// the input filter string
		1,				// 1 = generate an optimized BPF filter
		0xFFFFFF00);	// use the 255.255.255.0 netmask

	if (result != 0)
	{
		printf("Error compiling the filter %s\n", pcap_geterr(pPcap));
		pcap_close(pPcap);
		return -1;
	}

	//
	// set the filter
	//
	result = pcap_setfilter(pPcap, &bpfProgram);

	if (result != 0)
	{
		printf("Error setting the filter %s\n", pcap_geterr(pPcap));
		pcap_close(pPcap);
		return -1;
	}

	//
	//receive the packets and process them
	//
	do
	{
		if (_kbhit())
		{
			//
			// the user wants to exit
			//
			// NOTE: be careful not to use this approach in production code.
			// kbhit() is an expensive operation if called for every received packet.
			//
			// call it
			// every X received packets or
			// from within the if (result == 0){} code i.e. if a timeout occurs
			//
			break;
		}

		result = pcap_next_ex(pPcap, &pPcapHeader, &pData);

		if (result < 0)
		{
			printf("Error receiving packets: %s\n", pcap_geterr(pPcap));
			//
			// exit from the loop
			//
			break;
		}

		if (result == 0)
		{
			//
			// timeout elapsed and no packets where received
			//
			continue;
		}

		//
		// we have a valid packet, try to extract the mac address and check if it's IP
		//
		DissectPacket(pPcapHeader, pData);

	}while(TRUE);

	pcap_freecode(&bpfProgram);
	pcap_close(pPcap);

	return 0;
}

static
VOID DissectPacket(struct pcap_pkthdr *pPcapHeader, const u_char *pData)
{
	USHORT etherType;
	PIP_HEADER pIpHeader;

	// | MAC dst (6 bytes) | MAC src (6 bytes) | ethertype/len (2 bytes) |

	//
	// check that the captured packet is long enough to keep the dest MAC
	//
	if (pPcapHeader->caplen >= 6)
	{
		printf("MAC Destination = %02x:%02x:%02x:%02x:%02x:%02x\n",
			pData[0],
			pData[1],
			pData[2],
			pData[3],
			pData[4],
			pData[5]);
	}
	else
	{
		printf("Truncated packet\n");
	}

	//
	// check that the captured packet is long enough to keep the src MAC
	//
	if (pPcapHeader->caplen >= 12)
	{
		printf("MAC Source = %02x:%02x:%02x:%02x:%02x:%02x\n",
			pData[6],
			pData[7],
			pData[8],
			pData[9],
			pData[10],
			pData[11]);
	}
	else
	{
		printf("Truncated packet\n");
	}

	//
	// check that the captured packet is long enough to keep the src MAC
	//
	if (pPcapHeader->caplen >= 14)
	{
		// 
		// extract the ethertype and convert it to host byte order
		//
		// we need to add ws2_32.lib in the lib dependencies to make use
		// of ntohs
		//
		etherType = ntohs(*(u_short*)&pData[12]);
		
		printf("Ethertype = 0x%04x\n", etherType);
	}
	else
	{
		printf("Truncated packet\n");
		//
		// stop the dissection
		//
		return;
	}
	
	//
	// check if it's IP and continue dissection
	//
	if (etherType == 0x0800)
	{
		printf("Packet is IP\n");

		//
		// let's check if we can cast it to an IP header
		//
		if (pPcapHeader->caplen <= 14 + sizeof(IP_HEADER))
		{
			printf("Truncated IP header\n");
			//
			// stop the dissection
			//
			return;
		}

		pIpHeader = (PIP_HEADER)(&pData[14]);

		//
		// print some fields. 
		//

		//
		// the header length is in number of 4bytes words, and it's saved
		// in the least significant 4 bits of the first byte of the packet
		// It does *NOT* coincide with sizeof(IP_HEADER) when there are
		// options in the IP header
		//
		printf("The header length is %u\n", (pIpHeader->VerIHL & 0x0f) * 4);

		//
		// the length is in network byte order, be sure to convert it to host byte order
		//
		printf("The total length of the IP packet is %u bytes\n", ntohs(pIpHeader->TotalLength));

		printf("Encapsulated protocol: ");
		switch(pIpHeader->Protocol)
		{
		case 0x01: printf("ICMP"); break;
		case 0x06: printf("TCP"); break;
		case 0x11: printf("UDP"); break;
		default: printf("unknown (%u)", pIpHeader->Protocol);
		}
		printf("\n");
	}
}
